###Nearest singular polynomial matrix helper functions
####Author: Joseph Haraldson
###Purpose: Utility functions used to compute the nearest singular structured matrix.
###NOTE: Some of these functions are not used in the release implementation. 
###FULLY DOCUMENTED CODE AVAILABLE UPON REQUEST. Email [jharalds] [at] [uwaterloo.ca] to request the documented version.
### Cite this code with a reference to the paper it was written for: 
### "Computing the Nearest Rank-Deficient Matrix Polynomial by Haraldson et al. '17"
### All rights reserved.

generate_structured_ansatz := proc(A,x:='xx',symbol:='a')
local M:=Matrix(A);
local vars;
local i,j,k;
vars:=[];
for i from 1 to LinearAlgebra:-RowDimension(A) do 
  for j from 1 to LinearAlgebra:-ColumnDimension(A) do
    M[i,j]:= add(symbol[i,j,k]*x^k,k=0..degree(A[i,j],x));
    vars:= [op(vars),seq(symbol[i,j,k],k=0..degree(A[i,j],x))];
  od:
od:

return M,vars;
end proc:

generate_ansatz := proc(n,d,x,symbol:='a')
local M:=Matrix(n);
local vars;
local i,j;
vars:=[];
for i from 1 to n do 
  for j from 1 to n do 
       M[i,j] := add(symbol[i,j,k]*x^k,k=0..d);
       vars:= [op(vars),seq(symbol[i,j,k],k=0..d)];
  od:
od:

return M,vars;
end proc:


norm_opt:=proc(f,x,p:=2)
local v;
local nm;
v:=PolynomialTools:-CoefficientVector(f,x);

nm := add(v[i]^p,i=1..ArrayNumElems(v));

return (nm)^(1/p);
end proc:


poly_norm_opt :=proc(A,x,p:=2)

local nm;
local i;

nm := add(norm_opt(A(i),x,2)^2, i=1..ArrayNumElems(A));
return sqrt(nm);

end proc:






generate_factor :=proc(u,t,d,n,m,structure:=NULL)
  local 	my_vars:=[], 
		U, 
		i,
		j;
		
	U:=Matrix(n,m):
	my_vars := []:
	
	
	
	for i from 1 to n do
	  for j from 1 to m do 
	      if(structure <> NULL) then
		  U[i,j] := add(u[i, j, k]*t^k, k = 0 .. degree(structure[i, j], t));
		  my_vars:= [op(my_vars), seq(u[i, j, k], k = 0 .. degree(structure[i,j],t))];
	      else
		  U[i,j] := add(u[i, j, k]*t^k, k = 0 .. d);
		  my_vars:= [op(my_vars), seq(u[i, j, k], k = 0 .. d)];      
	      fi:
	  od:
	od:
	
	return U, my_vars
end proc;

generate_triangular_factor :=proc(u,t,d,n,m,unit:=true)
  local 	my_vars:=[], 
		U, 
		i,
		j;
		
	U:=Matrix(n,m):
	my_vars := []:
	
	
	
	for i from 1 to n do
	  for j from 1 to m do 
	      if i > j then 
		  U[i,j] := add(u[i, j, k]*t^k, k = 0 .. d);
		  my_vars:= [op(my_vars), seq(u[i, j, k], k = 0 .. d)];
	      elif i= j then
		  U[i,j]:=1;
	      fi:
	  od:
	od:
	
	return U, my_vars
end proc;


clean_up := proc (A, t, tol:=1e-4)

    local i,
	  j,
	  cf_vec
	  ;

	  for i from 1 to ArrayNumElems(A) do 
	      cf_vec:= PolynomialTools:-CoefficientVector(A(i),t);
	      for j from 1 to ArrayNumElems(cf_vec) do 
		  if abs(cf_vec(j)) < tol then 
		  cf_vec(j) := 0;
		  fi:
	      od:
	      A(i):= PolynomialTools:-FromCoefficientVector(cf_vec,t);
	  od:
	  
	  return A;
	  
end proc:
clean_list :=proc (L, tol:=1e-6)
    local i,
	  M
	  ;
	  M:=[]:
	  for i from 1 to nops(L) do 
	      
	      if abs(rhs(L[i])) < tol then 
		M:=[op(M), lhs(L[i]) = 0];
		  
	      else
		 M:=[op(M), L[i]];
	      fi:
	  od:
    
    return M;
end proc:



matrix_degree := proc( A,t)

    local i,
	  d;
	
    d := degree(A(1),t);  
    for i from 2 to ArrayNumElems(A) do 
	d := max(degree(A(i),t),d);
    od:
    return d;
end proc:

nearest_unstructured_distance := proc(A,t,tol:=1e-8)
    local i,
	  j,
	  bb,
	  bb_vars,
	  d,
	  n,
	  m,
	  my_obj,
	  my_eqns,
	  cf_vec,
	  Ab,
	  AA,
	  mu,
	  Sigma,
	  rank
	  ;
	  
	  d := matrix_degree (A,t);
	  n,m := LinearAlgebra:-Dimension(A);

	  bb := Vector(m, i-> add(b[i,j]*t^j,j=0..2*max(m,n)*d));
	  bb_vars := [seq(seq(b[i,j],j=0..2*max(m,n)*d),i=1..m)];
	  
	  mu := 2*n*d;

	  Ab := simplify(A.bb);   
	  my_eqns := []:
	  for i from 1 to ArrayNumElems(Ab) do 
	      cf_vec := PolynomialTools:-CoefficientVector(Ab(i),t);
	      my_eqns  := [op(my_eqns), op(Equate(cf_vec, Vector(ArrayNumElems(cf_vec)))) ];
	  od:
	  
	  AA:=LinearAlgebra:-GenerateMatrix(my_eqns,bb_vars)[1];
	  
	  Sigma:= LinearAlgebra:-SingularValues(evalf(AA)); 
	  rank:=0;
	  for j from 1 to ArrayNumElems(Sigma) do 
	      if abs(Sigma[j]) < tol then break fi:
	  od:
	  
	  rank:= ceil(evalf(j-1)/evalf(mu+d+1));
	  
	  
	  return [rank, Sigma[j-1],j];
end proc:




convolution_matrix := proc(a,d,n,t)
      local mat,
	    i,
	    j,
	    mu,
	    cf_vec
	    ;
  
	  mu:= n*d+1;
	  mat  := Matrix(mu,mu+d);
	  
	  for i from 1 to mu do 
	      mat [i] := PolynomialTools:-CoefficientVector(a*t^(i-1),t);
	  od:
	  
	 return LinearAlgebra:-Transpose(mat);
end proc:




linearize_A_fat := proc(A,t)

local 	d,
	n,
	m,
	bb,
	bb_vars,
	Ab,
	my_eqns,
	cf_Vec,
	AA,
	b_lin,
	mu,
	i
	;


	d:= matrix_degree(A,t);
	n,m := LinearAlgebra:-Dimension(A);
	
	#WARNING: assumes square matrix, i.e. m=n
	AA:=`<,>`(seq(`<|>`(seq(convolution_matrix(A[i, j], d, n, t), j = 1 .. n)), i = 1 .. n));
	

	return AA;
	
end proc:



linearize_A := proc(A,t)
    local d,
	  n,
	  m,
	  bb,
	  bb_vars,
	  Ab,
	  my_eqns,
	  cf_vec,
	  AA,
	  b_lin,
	  mu,
	  i
	  ;
    
    	  d := matrix_degree (A,t); 
	  n,m := LinearAlgebra:-Dimension(A);

	  mu:=1*min(m,n)*d+1;
	  
	  bb := Vector(m, i-> add('b'[i,j]*t^j,j=0..min(m,n)*d));
	  bb_vars := [seq(seq('b'[i,j],j=0..mu-1),i=1..m)];

	  
	  Ab := (A.bb); 
	  my_eqns := []:
	  for i from 1 to ArrayNumElems(Ab) do 
	      cf_vec := PolynomialTools:-CoefficientVector(Ab(i),t);
	      my_eqns  := [op(my_eqns), op(Equate(cf_vec, Vector(ArrayNumElems(cf_vec)))) ];
	  od:
	  
	  AA:=LinearAlgebra:-GenerateMatrix(my_eqns,bb_vars)[1];

	  
	  return AA, convert(bb_vars,Vector);
end proc:



linearize_A_minimal := proc(A,t,b_guess)
    local d,
	  n,
	  m,
	  bb,
	  bb_vars,
	  Ab,
	  my_eqns,
	  cf_vec,
	  AA,
	  b_lin,
	  mu,
	  i,
	  k
	  ;
    
    	  d := matrix_degree (A,t); 
	  n,m := LinearAlgebra:-Dimension(A);

	  mu:=1*min(m,n)*d+1;
	  
	  bb_vars := []:
	  bb:=Vector(m);
	  
	  
	   for i from 1 to m do   
		cf_vec := PolynomialTools:-CoefficientVector(b_guess[i],t);
		
		bb_vars := [op(bb_vars), seq(b[i,k], k=0..degree(b_guess[i],t))];
		bb[i] := add(b[i,k]*t^k,k=0..degree(b_guess[i],t));
		
		
		for k from 1 to ArrayNumElems(cf_vec) do
		    if(cf_vec[k]=0) then
			bb_vars := convert(convert(bb_vars,set) minus {b[i,k-1]},list);
			bb[i]:=subs(b[i,k-1]=0,bb[i]);
		    fi:
		od:
	      od:
	 
	
	  
	  Ab := (A.bb); 
	  
	  my_eqns := []:
	  for i from 1 to ArrayNumElems(Ab) do 
	      cf_vec := PolynomialTools:-CoefficientVector(Ab(i),t);
	      my_eqns  := [op(my_eqns), op(Equate(cf_vec, Vector(ArrayNumElems(cf_vec)))) ];
	  od:
	 AA:=LinearAlgebra:-GenerateMatrix(my_eqns,bb_vars)[1];

	  
	  return AA, convert(bb_vars,Vector);
end proc:


rank_factor_A_lin := proc(A_lin,r)
  local 	 U_lin,
		 V_lin,
		 U_lin_vars,
		 V_lin_vars,
		 n,
		 m,
		 i,
		 j
		 ;


	n,m := LinearAlgebra:-Dimension(A_lin);
	U_lin := Matrix(n,r);
	V_lin := Matrix(r,m);
	
	for i from 1 to n do
	    for j from 1 to r do 
		U_lin[i,j] := 'U'[i,j];
	    od:
	od:
	
	
	for i from 1 to r do 
	    for j from 1 to m do 
		 V_lin[i,j] := 'V'[i,j];
	    od:
	od:
	
	U_lin_vars := [seq( seq('U'[i,j],j=1..r),i=1..n)];
	V_lin_vars := [seq( seq('V'[i,j],j=1..m),i=1..r)];

	
	return U_lin,V_lin, U_lin_vars,V_lin_vars;
end proc:

enforce_polynomial_constraints := proc( A_lin, U_lin, V_lin,t)
    local 	i,
		j,
		constraints,
		UV
		;

	constraints := [];
	
	UV := simplify(U_lin.V_lin);
	
	for i from 1 to ArrayNumElems(A_lin) do 
	    if A_lin(i) =0 then 
		constraints:=[op(constraints), UV(i)=0]:
	    fi:
	od:
	
	return constraints;
end proc:

random_pt := proc(L)
    local i, M := []:
    
    for i from 1 to nops(L) do 
	M:= [op(M), L[i] = RandomTools:-MersenneTwister:-GenerateFloat()];
    od:
    return M;
end proc:





generate_structured_perturbation := proc(A,t,symbol:='a')
	local i,
	      M,
	      j,
	      m,
	      n,
	      k,
	      a_vars,
	      cf_vec;
	      ;
	      
	      
	      n,m := LinearAlgebra:-Dimension(A);
	      M:= Matrix(n,m);
	      a_vars := []:
	      
	      for i from 1 to n do 
		  for j from 1 to m do 
			cf_vec := PolynomialTools:-CoefficientVector(A[i,j],t);
			
			a_vars := [op(a_vars), seq(symbol[i,j,k], k=0..degree(A[i,j],t))];
			M[i,j] := add(symbol[i,j,k]*t^k,k=0..degree(A[i,j],t));
			
			
			for k from 1 to ArrayNumElems(cf_vec) do
			    if(cf_vec[k]=0) then
				a_vars := convert(convert(a_vars,set) minus convert(symbol[i,j,k-1],set),list);
				M[i,j]:=subs(symbol[i,j,k-1]=0,M[i,j]);
			    fi:
			od:
			
		  od:
	      od:
	      
	      
	      return M, a_vars;
end proc:

vectorize := proc(b,t)
    local i,
	  j,
	  vec,
	  ell
	  ;
	  
	  ell := []:
	  for i from 1 to ArrayNumElems(b) do 
	      vec := PolynomialTools:-CoefficientVector(b(i),t);
	      for j from 1 to ArrayNumElems(vec) do 
		  if(vec[j] <>0) then
		      ell := [op(ell),vec[j]];
		  fi:
	      od:
	  od:
	  
	  return convert(ell,Vector);

end proc:


constrained_newton := proc (A,DeltaA,DeltaA_vars, A_init,b_init,num_iter := 1000,tol:=1e-9,divergence:=1e2)
	local A_vars, 
	      b_vars, 
	      Lambda, 
	      Lambda_init, 
	      A_lin, 
	      A_init_lin,
	      b_lin, 
	      DA_lin, 
	      X, 
	      X_init,
	      Z, 
	      Z_init,
	      ZZ_next,
	      DeltaZ, 
	      ZZ,
	      residuals, 
	      L, 
	      Phi, 
	      H, 
	      G,
	      Gx,
	      Gy,
	      lagrangian_system,
	      cf_vec,
	      cf_mat,
	      ell,
	      default_DeltaA_pt,
	      i,
	      active_vars,
	      J,
	      ZZ_prev,
	      my_eqns,
	      ls_soln,
	      unactive_vars,
	      trivial_eqns
	      ;
	      
	      
	     A_lin := linearize_A_minimal(A,t,b_init)[1];
	     A_init_lin := linearize_A_minimal(A_init,t,b_init)[1];
	     DA_lin,b_lin := linearize_A_minimal(DeltaA,t,b_init);
	     
	     
	    
	     b_vars:= convert(b_lin,list);
	   
	    default_DeltaA_pt := [seq(DeltaA_vars[i]=0,i=1..nops(DeltaA_vars))];


		
	   print(A_init);
	    
		
	     X_init :=[ op(Equate(DA_lin, A_init_lin-A_lin)), op(Equate(b_lin, vectorize(b_init,t)))];
	     
	    
	     
	     residuals := convert( (A_lin + DA_lin).b_lin,list);
	     residuals := convert( [op(residuals), poly_norm_opt(b_lin,t)^2-1],set) minus {0};
	     residuals := convert(residuals,list); 
	     
	     active_vars:= convert(convert([seq( lhs(X_init[i]),i=1..nops(X_init))],set) minus {0,0.0},list);
	     
	     unactive_vars := convert(convert(DeltaA_vars,set) minus convert(active_vars,set),list);
	     trivial_eqns := [seq(unactive_vars[i]=0,i=1..nops(unactive_vars))];
	     Lambda := [seq('lambda'[i],i=1..nops(residuals))];
	     X:=active_vars ; 	     Z:= [op(X),op(Lambda)];
	   

	     Phi:= poly_norm_opt(DeltaA,t)^2 +(poly_norm_opt(b_lin,t)^2-1);
	     L:=Phi - LinearAlgebra:-DotProduct(convert(Lambda,Vector), convert(residuals,Vector), conjugate=false);
	     
	
	     G:= VectorCalculus:-Gradient(L,Z);
	     H:=VectorCalculus:-Hessian(L,Z);
	     Gx := (VectorCalculus:-Gradient(L,X));
	     Gy := (VectorCalculus:-Gradient(L,Lambda));
	     
	     
	     
	      J:= VectorCalculus:-Jacobian(residuals,X);
	     if( LinearAlgebra:-Rank(subs(X_init,J)) <> LinearAlgebra:-RowDimension(J)) then
		WARNING("RANK DEFICIENT JACOBIAN DETECTED. PREPARE FOR CRITICAL FAIL");
		WARNING("Possibly not a minimally embedded nullspace vector");
	     fi:
     
	     
	     lagrangian_system := subs(default_DeltaA_pt, convert(subs(X_init,Gx),Vector));
	     my_eqns := Equate(lagrangian_system, Vector(ArrayNumElems(lagrangian_system)));
	     cf_mat,cf_vec := LinearAlgebra:-GenerateMatrix(my_eqns,Lambda);
	     ls_soln := Optimization:-LSSolve([cf_vec,cf_mat])[2];
	     
	     Lambda_init := Equate(convert(Lambda,Vector), ls_soln);
	     Z_init := [op(X_init), op(Lambda_init)];
	     

	     
	     ZZ:= subs(default_DeltaA_pt, convert(subs(Z_init,Z),Vector));
	     ZZ_next := Equate(Z,ZZ);
	     for i from 1 to num_iter do
		DeltaZ:= Optimization:-LSSolve([subs(ZZ_next,-1*G)  , subs(ZZ_next, H)])[2];
		ZZ_prev := ZZ;
		ZZ:= ZZ+DeltaZ; 
		ZZ_next := Equate(Z,ZZ);

		print(norm(ZZ-ZZ_prev,2));
		if norm(ZZ-ZZ_prev,2) < tol then break fi:
		
		if norm(ZZ-ZZ_prev,2) > divergence then WARNING("Newton Diverged"); break; fi:

	     od:
	     
	     
	     
	     return [subs( default_DeltaA_pt, subs(ZZ_next,Phi)),[ op(ZZ_next), op(trivial_eqns)],i,norm(ZZ-ZZ_prev,2)];

end proc:


generate_initial_point := proc(A,A_init,DA,t)
	local i,
	      j,
	      n,
	      cf_vec,
	      perturb_vec; 
	      eqns
	      ;
	      
	      eqns:=[]:
	      for i from 1 to ArrayNumElems(A) do 
		    cf_vec := PolynomialTools:-CoefficientVector(A(i)-A_init(i),t);
		    perturb_vec := PolynomialTools:-CoefficientVector(DA(i),t);
		    cf_vec := Vector(cf_vec, ArrayNumElems(perturb_vec));
		    eqns:= [op(eqns), op(Equate(perturb_vec,cf_vec))];
	      od:
	 return eqns;     
end proc:

#this function is for arbitrary linear structures... it's not for the paper. just for me.
#perturbation structure specified by the user.
#it is not to be considered robust.
constrained_newton_arbitrary := proc (A,DeltaA,DeltaA_vars, A_init,b_init,num_iter := 1000,tol:=1e-9,divergence:=1e2)
	local A_vars, 
	      b_vars,
	      Lambda,
	      Lambda_init,
	      A_lin,
	      A_init_lin,
	      b_lin,
	      DA_lin,
	      X,
	      X_init,
	      Z, 
	      Z_init,
	      ZZ_next,
	      DeltaZ, 
	      ZZ,  
	      residuals,
	      L, 
	      Phi,
	      H, 
	      G, 
	      Gx,
	      Gy,
	      lagrangian_system,
	      cf_vec,
	      cf_mat,
	      ell,
	      default_DeltaA_pt,
	      i,
	      active_vars,
	      J,
	      ZZ_prev,
	      my_eqns,
	      ls_soln,
	      unactive_vars,
	      trivial_eqns
	      ;
	      
	     A_lin := linearize_A_minimal(A+DeltaA,t,b_init)[1];
	     A_init_lin := linearize_A_minimal(A_init+DeltaA,t,b_init)[1];
	     DA_lin,b_lin := linearize_A_minimal(DeltaA,t,b_init);
	     
	     
	    
	     b_vars:= convert(b_lin,list);
	    
	    default_DeltaA_pt := [seq(DeltaA_vars[i]=0,i=1..nops(DeltaA_vars))];
	    
	    A_lin := subs(default_DeltaA_pt,A_lin); 
	    A_init_lin := subs(default_DeltaA_pt,A_init_lin); 
		
	  
		
	     X_init :=[ op(Equate(DA_lin, A_init_lin-A_lin)), op(Equate(b_lin, vectorize(b_init,t)))];
	     residuals := convert( (A_lin + DA_lin).b_lin,list);
	     residuals := convert( [op(residuals), poly_norm_opt(b_lin,t)^2-1],set) minus {0}; 
	     residuals := convert(residuals,list); 
	     
	     
	     active_vars:= convert(convert([seq( lhs(X_init[i]),i=1..nops(X_init))],set) minus {0,0.0},list);
	     
	     unactive_vars := convert(convert(DeltaA_vars,set) minus convert(active_vars,set),list);
	     trivial_eqns := [seq(unactive_vars[i]=0,i=1..nops(unactive_vars))];
	  
	     Lambda := [seq('lambda'[i],i=1..nops(residuals))];
	     X:=active_vars ; 
	     Z:= [op(X),op(Lambda)];
	   

	     Phi:= poly_norm_opt(DeltaA,t)^2 +(poly_norm_opt(b_lin,t)^2-1);
	     L:=Phi - LinearAlgebra:-DotProduct(convert(Lambda,Vector), convert(residuals,Vector), conjugate=false);
	    
	     G:= VectorCalculus:-Gradient(L,Z);
	     H:=VectorCalculus:-Hessian(L,Z);
	     Gx := (VectorCalculus:-Gradient(L,X));
	     Gy := (VectorCalculus:-Gradient(L,Lambda));
	     
	     
	     
	      J:= VectorCalculus:-Jacobian(residuals,X);
	     if( LinearAlgebra:-Rank(subs(X_init,J)) <> LinearAlgebra:-RowDimension(J)) then
		WARNING("RANK DEFICIENT JACOBIAN DETECTED. PREPARE FOR CRITICAL FAIL");
		WARNING("Possibly not a minimally embedded nullspace vector");
	     fi:
	     lagrangian_system := subs(default_DeltaA_pt, convert(subs(X_init,Gx),Vector));
	     my_eqns := Equate(lagrangian_system, Vector(ArrayNumElems(lagrangian_system)));
	     cf_mat,cf_vec := LinearAlgebra:-GenerateMatrix(my_eqns,Lambda);
	     ls_soln := Optimization:-LSSolve([cf_vec,cf_mat])[2];
	     
	     Lambda_init := Equate(convert(Lambda,Vector), ls_soln);
	     Z_init := [op(X_init), op(Lambda_init)];
	     
	     ZZ:= subs(default_DeltaA_pt, convert(subs(Z_init,Z),Vector));
	     ZZ_next := Equate(Z,ZZ);
	     for i from 1 to num_iter do
		DeltaZ:= Optimization:-LSSolve([subs(ZZ_next,-1*G)  , subs(ZZ_next, H)])[2];
		ZZ_prev := ZZ;
		ZZ:= ZZ+DeltaZ; 		ZZ_next := Equate(Z,ZZ);

		print(norm(ZZ-ZZ_prev,2));
		if norm(ZZ-ZZ_prev,2) < tol then break fi:
		
		if norm(ZZ-ZZ_prev,2) > divergence then WARNING("Newton Diverged"); break; fi:

	     od:
	     
	     
	     
	     return [subs( default_DeltaA_pt, subs(ZZ_next,Phi)),[ op(ZZ_next), op(trivial_eqns)],i,norm(ZZ-ZZ_prev,2)];

end proc:

generate_singular_matrix:= proc( n,m, r,d,t,structure:=dense,lower:=-10,upper:=10)
    local i,
	  j,
	  Ml,
	  Mr
	  ;


	  Ml := Matrix(n,r);
	  Mr :=  Matrix(r,m);
	  
      for i from 1 to n do 
	  for j from 1 to  r do 
	      Ml [i,j] := randpoly(t,degree= floor(d/2.0),coeffs=rand(lower..upper),structure);
	  od:
      od:

      for i from 1 to r do 
	  for j from 1 to m do 
	      Mr [i,j] := randpoly(t,degree= ceil(d/2.0), coeffs=rand(lower..upper),structure);
	  od:
      od:
      return simplify(Ml.Mr);
end proc:


generate_singular_matrix_rank_factored:= proc( n,m, r,d,t,structure:=dense,lower:=-10,upper:=10)
    local i,
	  j,
	  Ml,
	  Mr
	  ;


	  Ml := Matrix(n,r);
	  Mr :=  Matrix(r,m);
	  
      for i from 1 to n do 
	  for j from 1 to  r do 
	      Ml [i,j] := randpoly(t,degree= floor(d/2.0),coeffs=rand(lower..upper),structure);
	  od:
      od:

      for i from 1 to r do 
	  for j from 1 to m do 
	      Mr [i,j] := randpoly(t,degree= ceil(d/2.0), coeffs=rand(lower..upper),structure);
	  od:
      od:
      return simplify(Ml.Mr),Ml,Mr;
end proc:



clear_denoms := proc(b_input,t)

      local i,
	    bb
	    ;
  
	    
	    bb := simplify((b_input)); 
	    for i from 1 to ArrayNumElems(bb) do 
		  bb := simplify(bb * denom(bb(i)));
	    od:
	

      return bb;

end proc:


generate_unstructured_perturbation := proc(A,t,symbol:='a')

      local i,
	    M,
	    j,
	    a_vars
	    ;
	    
	    M:= Matrix(LinearAlgebra:-Dimension(A));
	    a_vars := []:
	    
	    for i from 1 to LinearAlgebra:-RowDimension(A) do
		for j from 1 to LinearAlgebra:-ColumnDimension(A) do 
		    M[i,j] := add(symbol[i,j,k]*t^k, k=0..degree(A[i,j],t));
		    a_vars := [op(a_vars), seq(symbol[i,j,k], k=0..degree(A[i,j],t))];
		od:
	    od:

	    return M,a_vars;
end proc:


generate_dense_unstructured_perturbation:= proc(A,t, symbol:='a')
      local i,
	    M,
	    j,
	    d,
	    a_vars
	    ;
	    
	    M:= Matrix(LinearAlgebra:-Dimension(A));
	    a_vars := []:
	    d:= matrix_degree(A,t);
	    
	    for i from 1 to LinearAlgebra:-RowDimension(A) do
		for j from 1 to LinearAlgebra:-ColumnDimension(A) do 
		    M[i,j] := add(symbol[i,j,k]*t^k, k=0..d);
		    a_vars := [op(a_vars), seq(symbol[i,j,k], k=0..d)];
		od:
	    od:

	    return M,a_vars;
end proc:

generate_perturbation := proc(A,magnitude:=1)
      local DA,
	    DA_vars,
	    random_pt
	    ;
	    
	    
	    DA,DA_vars:=generate_structured_perturbation(A,t);
	    random_pt := [seq(DA_vars[i]= (-1)^(RandomTools:-MersenneTwister:-GenerateInteger())*RandomTools:-MersenneTwister:-GenerateFloat()*magnitude , i=1..nops(DA_vars))];
	   
	    return subs(random_pt, DA);
	    
end proc:



make_primitive := proc(b,t)
	local c,
	      i,
	      cont
	      ;
	     
	      c:=Vector(ArrayNumElems(b));
	      cont := b(1):
	      for i from 2 to ArrayNumElems(b) do 
		  cont := gcd( cont,b(i));
	      od:

	      for i from 1 to ArrayNumElems(b) do 
		  c[i] := quo(b(i),cont,t);
		  if rem(b(i),cont,t) <> 0 then 
		      WARNING("FAILED TO COMPUTE GCD");
		  fi:
	      od:
	      
	      return c;
end proc:



check_support := proc(A,B,t)
    local i;
    
    for i from 1 to ArrayNumElems(A) do 
	if degree(A(i),t) <> degree(B(i),t) then 
	  return false; 
	fi:
    od:
  return true;
end proc:

