#Author: Joseph Haraldson
#some code to perform numeric polynomial linear algebra. 
#depends on the nearest_singular_poly code base for some basic linearization functions.
###NOTE: Some of these functions are not used in the release implementation. 
###FULLY DOCUMENTED CODE AVAILABLE UPON REQUEST. Email [jharalds] [at] [uwaterloo.ca] to request the documented version.
### Cite this code with a reference to the paper it was written for: 
### "Computing the Nearest Rank-Deficient Matrix Polynomial by Haraldson et al. '17"
### All rights reserved.

#preliminary verison

clean_up_noise:=proc(A,tol:=1e-4)
local i;
    for i from 1 to ArrayNumElems(A) do 
	if abs(A(i)) < tol then 
	    A(i):=0;
	fi:
    od:
    return A:
end proc:


numeric_rank_relative := proc(S, m, tol:=1e-4)
    local i,
	  r
	  ;
 
	  r:=m; 
	  for i from 1 to m-1 do 
	      if S[i+1]/S[i]  < tol * evalf(sqrt(m)) then 
		  r:= i+1;
	      fi:
	  od:
	  return r;
end proc:

numeric_rank_absolute := proc(S, m, tol:=1e-30)
    local i,
	  r
	  ;
 
	  r:=m; 
	  for i from 1 to m-1 do 
	     if S[i] < tol then
		r := i;
		break;
	     fi:
	  od:
	  return r;
end proc:



compute_orthonormal_kernel := proc(A_lin,tol:=1e-4)
    
    local  UU,
	  SS, 
	  VVt,
	  n,  
	  m, 
	  i,
	  singular_column_start_index,
	  singular_column_end_index
	  ;
    

    
    UU,SS,VVt :=LinearAlgebra:-SingularValues(evalf(A_lin),output = ['U','S','Vt']); 

    n,m := LinearAlgebra:-Dimension(A_lin);
    singular_column_start_index := 1;
    singular_column_end_index := m; 
    
    singular_column_start_index := numeric_rank_absolute(SS,m,tol);
  
    return LinearAlgebra:-Transpose(VVt[singular_column_start_index..singular_column_end_index]);
end proc:



#procedure to compute a minimally embedded kernel vector
#column basis K of the kernel of a (scalar)
#WARNING: what we really want is a LQ factorization and not an RQ factorization
#THIS PERFORMS AN RQ FACTORIZATION
#THIS DOES NOT GIVE A MINIMALLY DEGREE EMBEDDED VECTOR
#THE RESULTING VECTOR WILL NOT BE PRIMITIVE IF IT CORRESPONDS TO A MATRIX POLYNOMIAL
#HOWEVER, IT WILL BE MINIMALLY EMBEEDDED (it is a t^k multiple of a minimally degree embedded vector).
#THIS WARNING IS IMPORTANT.
extract_minimally_embedded_kernel_vector := proc(K)
    local Q, 
	  R,
	  n
	  ;

	 Q,R := LinearAlgebra:-QRDecomposition(LinearAlgebra:-Transpose(K));
	  
	  return LinearAlgebra:-Transpose(R[LinearAlgebra:-RowDimension(R)]);
end proc:



 
devectorize := proc(b,n,d,t,tol:=1e-6)
local 	bb,
	bb_vars,
	i,
	j
	;
	


	bb := Vector(n, i-> add('c'[i,j] *t^j, j=0..n*d));


	bb_vec := vectorize(bb,t);
	my_eqns := Equate(bb_vec, clean_up_noise(b,tol));
	
	return (subs(my_eqns, bb)); #devectorization complete
end proc:



